// Title:   Action.h
// Created: Thu Sep 25 15:17:01 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __ACTION_H
#define __ACTION_H

/**
 * This file contains some (more or less) abstract classes that mark specific actions.
 *
 * @author Simon Heinzle
 * @version 1.0
 */

#include "../../../../DataTypes/src/Vector3D.h"
#include "BoxSelector.h"

/**
 * class Action
 *
 * This class defines a basic action with a name
 */
class Action {
protected:
	// Just a marker base class
	const char *_name;
public:
	/**
	 * Constructor
	 */
	Action();
	/**
	 * Returns the name of this class
	 *
	 * @return pointer to <code>char</code> containing the name
	 */
	const char* getName() const;
};

/**
 * class UndoAction
 *
 * This class defines an undo operation
 */
class UndoAction:public Action {
public:
	/** 
	 * Constructor
	 */
	UndoAction();
	/**
	 * Deconstructor
	 */
	virtual ~UndoAction() {};
	/** 
	 * Performs an undo operation and returns an inverse operation
	 * to this, for redo.
	 *
	 * @return inverse operation <code>UndoAction</code> to this operation.
	 */
	virtual UndoAction* undo() = 0;
};

/**
 * class NormalAction
 *
 * this defines a normal operation on a selection
 */
class NormalAction:public Action {
protected:
	int _selection_index;
public:
	/**
	 * Constructor
	 */
	NormalAction();
	/**
	 * Executes the desired operation
	 */
	virtual void execute() = 0;
	/**
	 * Executes the desired operation and returns an inverse operation for undo.
	 *
	 * @return inverse operation <code>UndoAction</code> to this operation.
	 */
	virtual UndoAction* executeUndo() = 0;
	/**
	 * Sets the <code>int selection_index</code> in range [0..2]. Normally operations are performed
	 * only on the surfels selected by this flag (SELECTED1, SELECTED2, SELECTED3)
	 */
	void setSelection(int selection_index);
};

#endif

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
