// Title:   MLSScatterer.cpp
// Created: Thu Sep 25 15:16:37 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "MLSScatterer.h"
 
//#define _REAL_MLS_PROJECTION

#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include <qdatetime.h>
#include <qapplication.h>

#include "MovingLeastSquares.h"
#include "UndoRedoSpecialisations.h"
#include "Selector.h"

#include <gsl/gsl_multifit.h>
#include <gsl/gsl_errno.h>

#include "NewOverload.h"

MLSScatterer::MLSScatterer() {
	maxdiameter = 0.f;	
	_boxselector =0;
	_new_query_grid = 0;
	_use_mls = false;
	_param_mls = 0;
	_density_check = true;
	_boxselector = 0;
}

MLSScatterer::~MLSScatterer() {
	if (_new_query_grid) {
		_MCdelete((unsigned long)_new_query_grid);
		delete _new_query_grid;
	}
	if (_boxselector) {
		_MCdelete((unsigned long)_boxselector);
		delete _boxselector;
	}
}
	

void MLSScatterer::planeProjector(const Vector3D& source_position, int max_iterations, float min_error) {
	PlaneFit plane;
	float f;
	float epsilon;
	int current_iteration;
	Vector3D new_projected_source;
	
	_plane_projected_position = source_position;
	current_iteration = 0;
	do {
		plane.doPlaneFitAdaptiveWeightIncludeQueryGrid(_plane_projected_position, _neighbourhood_size, _new_query_grid);
		f = -1.f * plane.getDistance(_plane_projected_position);
		new_projected_source =  _plane_projected_position + plane.getNormal() * f;
		epsilon = Vector3D::distance(new_projected_source, _plane_projected_position) / _plane_projected_position.getLength();
		current_iteration++;
		_plane_projected_position = new_projected_source;
	} while ( (epsilon > min_error) && (current_iteration < max_iterations) );
	_planefit = plane;
}

void MLSScatterer::polynomialProjection(const Vector3D &source_position) {
	Selector *selector;
	KdTree *kdtree;
	// the vars for lsq fit
	gsl_matrix *X, *cov;
	gsl_vector *y,*c,*w;
	double chisquare;
	
	Vector3D u_direction;
	Vector3D v_direction;
	Vector3D normal;
	float z_coord;
	unsigned int i;

	std::vector<Vector3D> fit_positions;
	MinPriorityQueue<Vector3D, float> mpq;
	Vector3D pos;
	float param;
	float maxdist;

	
	normal = _planefit.getNormal();
	z_coord = -1.f / normal[2] * ( (normal[0] * 2 + normal[1] * 2 + _planefit.getD()) - (normal[0] + normal[1] + _planefit.getD()));
	u_direction = Vector3D(1.,1.,z_coord);
	v_direction = Vector3D(0.,0.,0.);
	
	u_direction.normalize();
	v_direction = Vector3D::crossProduct(u_direction, normal);
	v_direction.normalize();
	
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();

	// set up knext positions of source
	
	// build pqueue
	kdtree->setNOfNeighbours(_neighbourhood_size+1);
	kdtree->queryPosition(source_position);
	maxdist = (float) sqrt(kdtree->getSquaredDistance(_neighbourhood_size));
	
	_new_query_grid->setNOfNeighbours(_neighbourhood_size+1);
	_new_query_grid->queryAdvisoryAverageQueryRadius(maxdist);
	_new_query_grid->queryRange(source_position,maxdist);

	mpq.init();
	mpq.setSize(_neighbourhood_size + _new_query_grid->getNOfFoundNeighbours());
	// insert knext
	for (i=1; i<_neighbourhood_size; i++) {
		pos = kdtree->getNeighbourPosition(i);
		mpq.insert(pos,Vector3D::distance(pos,source_position));
	}
	// insert sfls from qggrid
	for (i=0; i<_new_query_grid->getNOfFoundNeighbours(); i++) {
		pos = _new_query_grid->getNeighbourPosition(i);
		mpq.insert(pos, Vector3D::distance(pos,source_position));
	}

	// get first k from pqueue
	for (i=0; i<_neighbourhood_size-1; i++) {
		fit_positions.push_back(mpq.getMinIndex());
		mpq.removeMin();
	}
	
	float distance_square = Vector3D::distance(fit_positions[_neighbourhood_size-2],source_position) * Vector3D::distance(fit_positions[_neighbourhood_size-2],source_position);
	param = distance_square / 3.0f;
	
	// setting up matrices for lsq fit of a function of order 3:
	// f(x,y) = a x^3 y^3 + b x^3 y^2 + c x^2 y^3 ...
	X = gsl_matrix_alloc(_neighbourhood_size-1,16);
	y = gsl_vector_alloc(_neighbourhood_size-1);
	w = gsl_vector_alloc(_neighbourhood_size-1);
	c = gsl_vector_alloc(16);
	cov = gsl_matrix_alloc(16,16);
	
	for (i = 0; i < _neighbourhood_size-1; i++) {
		Vector3D point = fit_positions[i];
		float fi = Vector3D::dotProduct(point - source_position, normal);
		float ui = Vector3D::dotProduct(point - source_position, u_direction);
		float vi = Vector3D::dotProduct(point - source_position, v_direction);

		gsl_matrix_set (X,  i, 0, ui*ui*ui*vi*vi*vi);
		gsl_matrix_set (X,  i, 1, ui*ui*ui*vi*vi);
		gsl_matrix_set (X,  i, 2, ui*ui*ui*vi);
		gsl_matrix_set (X,  i, 3, ui*ui*ui);
		gsl_matrix_set (X,  i, 4, ui*ui*vi*vi*vi);
		gsl_matrix_set (X,  i, 5, ui*ui*vi*vi);
		gsl_matrix_set (X,  i, 6, ui*ui*vi);
		gsl_matrix_set (X,  i, 7, ui*ui);
		gsl_matrix_set (X,  i, 8, ui*vi*vi*vi);
		gsl_matrix_set (X,  i, 9, ui*vi*vi);
		gsl_matrix_set (X,  i,10, ui*vi);
		gsl_matrix_set (X,  i,11, ui);
		gsl_matrix_set (X,  i,12, vi*vi*vi);
		gsl_matrix_set (X,  i,13, vi*vi);
		gsl_matrix_set (X,  i,14, vi);
		gsl_matrix_set (X,  i,15, 1);
		gsl_vector_set (y,  i, fi);
	
		float distance = Vector3D::distance(point, source_position);
		float weight = (float) exp(-(distance*distance)/param);
		gsl_vector_set (w, i,weight);
	}
	
	// solving least squares equation
	gsl_multifit_linear_workspace  *workspace = gsl_multifit_linear_alloc (_neighbourhood_size-1, 16);
	gsl_multifit_wlinear(X, w, y, c, cov, &chisquare, workspace);
	
	//_poly_projected_position = (source_position + _planefit.getNormal()*gsl_vector_get(c,15)) * _alpha + source_position * (1.f - _alpha);
	
	_poly_projected_position = source_position + normal * (float)gsl_vector_get(c,15); // * _alpha;
	gsl_matrix_free(X);
	gsl_matrix_free(cov);
	gsl_vector_free(y);
	gsl_vector_free(w);
	gsl_vector_free(c);
	gsl_multifit_linear_free(workspace);
}

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define MASK 123459876

float ran0(long *idum)	{
	long k;
	float ans;
	*idum ^= MASK; //XORing with MASK allows use of zero and other
				   //simple bit patterns for idum. 
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k; //Compute idum=(IA*idum) % IM without over-
								//.ows by Schrages method. 
	if (*idum < 0) *idum += IM;
	ans= (float) AM*(*idum); //Convert idum to a .oating result.
	*idum ^= MASK; //Unmask before return.
	return ans;
}

void MLSScatterer::getNewPositionsAxisAlignedBox(Vector3D min, Vector3D max) {
	float rand_x,rand_y, rand_z;
	QTime timer;
	Vector3D newpos;
	long *randval;
	
	timer = QTime::currentTime();
	randval = new long;
	_MCnew( (unsigned long)randval, sizeof(long), __FILE__, __LINE__);

	*randval = timer.msec() * timer.second()*timer.minute()*timer.hour() + 1;
	ran0(randval);

	maxdiameter = Vector3D::distance(min,max);
		
	if (_boxselector) {
		_boxselector = 0;
	}

	_new_positions.clear();
	for (unsigned int i=0; i<_no_of_points; i++) {
		// generate 3 random values
		rand_x = ran0(randval);//(float)rand() / (float)RAND_MAX;
		rand_y = ran0(randval);//(float)rand() / (float)RAND_MAX;
		rand_z = ran0(randval);//(float)rand() / (float)RAND_MAX;

		newpos[0] = rand_x * min[0] + (1.0f - rand_x) * max[0];
		newpos[1] = rand_y * min[1] + (1.0f - rand_y) * max[1];
		newpos[2] = rand_z * min[2] + (1.0f - rand_z) * max[2];
		
		_new_positions.push_back(newpos);
	}
	_MCdelete((unsigned long)randval);
	delete randval;

}

void MLSScatterer::getNewPositionsBox(BoxSelector *box_selector) {
	gsl_matrix *box_transformation_matrix;
	gsl_matrix *data_matrix;
	gsl_matrix *transformed_data_matrix;	
	unsigned int i;

	getNewPositionsAxisAlignedBox(Vector3D(-1.f,-1.f,-1.f), Vector3D(1.f,1.f,1.f));

	data_matrix = gsl_matrix_alloc(4,_new_positions.size());
	transformed_data_matrix = gsl_matrix_alloc(4,_new_positions.size());
	box_transformation_matrix = box_selector->getTransformationMatrixGSL();

	// fill data matrix
	for  (i=0; i<_new_positions.size(); i++) {
		gsl_matrix_set(data_matrix,0,i,(_new_positions[i])[0]);
		gsl_matrix_set(data_matrix,1,i,(_new_positions[i])[1]);
		gsl_matrix_set(data_matrix,2,i,(_new_positions[i])[2]);
		gsl_matrix_set(data_matrix,3,i,1);
	}
	// transform points
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, box_transformation_matrix, data_matrix, 0.0, transformed_data_matrix);

	// reset _new_positions
	for  (i=0; i<_new_positions.size(); i++) {
		(_new_positions[i])[0] = (float) gsl_matrix_get(transformed_data_matrix,0,i);
		(_new_positions[i])[1] = (float) gsl_matrix_get(transformed_data_matrix,1,i);
		(_new_positions[i])[2] = (float) gsl_matrix_get(transformed_data_matrix,2,i);
	}

	gsl_matrix_free(transformed_data_matrix);
	gsl_matrix_free(data_matrix);
	gsl_matrix_free(box_transformation_matrix);

}
	
void MLSScatterer::setMLSGauss(float param) {
	_param_mls = param;
}

void MLSScatterer::useMLS(bool use_it_or_loose_it) {
	_use_mls = use_it_or_loose_it;
}

void MLSScatterer::setDensityCheck(bool check_density) {
	_density_check = check_density;
}

void MLSScatterer::doSorting() {
	unsigned int i;
	KdTree *kdtree;
	Selector *selector;
	float dist;

	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();
	kdtree->setNOfNeighbours(5);
	
	_distance_pqueue.init();
	_distance_pqueue.setSize(_no_of_points);
	
	// set up distance priority queue
	for (i=0; i<_no_of_points; i++) {
		kdtree->queryPosition(_new_positions[i]);
		
		dist = kdtree->getSquaredDistance(1) + kdtree->getSquaredDistance(2)+ 
			   kdtree->getSquaredDistance(3) + kdtree->getSquaredDistance(4); 
		dist /= 4.0;

#ifdef _DEBUG
		if (_distance_pqueue.isFull())
			qDebug("Priority Queue is full!");
#endif
		_distance_pqueue.insert(i,dist);


	}

	

}

void MLSScatterer::setParameters(unsigned int no_of_points,unsigned int neighbourhood_size) {
	_no_of_points = no_of_points;
	_neighbourhood_size = neighbourhood_size;
}

void MLSScatterer::reColorSurfels(const std::vector<SurfelInterface*> &sfls,const std::vector<Vector3D> &poses) {
	Selector *selector;
	KdTree *kdtree;
	SurfelInterface *surfel;
	Scene *scene;
	Vector3D pos;

	QColor diffuse, specular;
	int rd,gd,bd;
	int rs,gs,bs;
	
	scene = Scene::getInstance();
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();

	for (unsigned int i=0; i<sfls.size(); i++) {
		kdtree->setNOfNeighbours(5);
		kdtree->queryPosition(poses[i]);
		
		rd = gd = bd = 0;
		rs = gs = bs = 0;
		
		for (int j=0; j<4; j++) {
			unsigned int real_index = kdtree->getNeighbourPositionIndex(j);
			surfel = selector->getSurfel(real_index);		
			diffuse = QColor(surfel->getDiffuseColor());
			specular = QColor(surfel->getSpecularColor());
			rd += diffuse.red();
			gd += diffuse.green();
			bd += diffuse.blue();
			rs += specular.red();
			gs += specular.green();
			bs += specular.blue();
		}
		
		rd /= 4;
		gd /= 4;
		bd /= 4;
		rs /= 4;
		gs /= 4;
		bs /= 4;
		diffuse.setRgb((int)rd,gd,bd);
		specular.setRgb(rs,bs,bs);

		scene->setSurfelDiffuseColor(sfls[i],diffuse.rgb(),false);
		scene->setSurfelSpecularColor(sfls[i], specular.rgb(),false);
	}
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
}


void MLSScatterer::execute() {
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	SurfelInterface *surfel;
	std::vector<SurfelInterface*> sfls;
	std::vector<Vector3D> poses;
	std::vector<Vector3D> *all_positions;
	MovingLeastSquares MLS(1e-3f, 30);
	unsigned int *index_array;
	unsigned i;
	unsigned index;

	Scene *scene;
	Selector *selector;
	KdTree *kdtree;

	MyDataTypes::TransformationMatrix16f trafo_matrix, inv_trafo_matrix;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selector->buildTree();
	kdtree = selector->returnKdTree();

	if ( !scene->getActiveObject() || !kdtree) {
		QApplication::restoreOverrideCursor();
		return ;
	}

	//get both transformationmatrix and its inverse
	scene->getActiveObject()->getTransformationMatrix(trafo_matrix);
	MtrInverse4x4f(trafo_matrix, inv_trafo_matrix);

	all_positions = selector->getAllPositionsWorldStdVec();
	index_array = new unsigned int[all_positions->size()];
	_MCnew((unsigned long)index_array, sizeof(index_array), __FILE__, __LINE__);
	
	kdtree->setNOfNeighbours(_neighbourhood_size);
	
	kdtree->queryPosition((*all_positions)[0]);
	
	MLS.setNOfNeighbours(_neighbourhood_size);
	MLS.setGaussParam(_param_mls);
	
	// set up query grid
	if (_new_query_grid) {
		_MCdelete((unsigned long)_new_query_grid);
		delete _new_query_grid;
	}
	
	_new_query_grid = new QueryGrid<uInt> (7, maxdiameter); // 7 used to be 5
	_MCnew( (unsigned long)_new_query_grid, sizeof(_new_query_grid), __FILE__, __LINE__);
	
	// sort points into pqueue
	doSorting();
	
	// first only plane projection
	for (i=0; i<_no_of_points; i++) {
		// get index with minimum length
		index = _distance_pqueue.getMinIndex();
		// remove index with minimum length
		_distance_pqueue.removeMin();

		// use this min for a MLS fit
		// rebuild mls after every 20 points
		if (i%20 == 0)
			MLS.setPositions(*all_positions);
		
		MLS.setSource(_new_positions[index]);

		if (_use_mls)
			MLS.doFullMLS();
		else
			MLS.doPlaneProjection();

		if (!MLS.isValid())
			continue;

		// check if we got a too high sampling density
		//  (not so sophisticated like in ~~> ParticleSimPlugin)
		float density = 0.f;
		float radii = 0.f;
		float count = 0.f;
		Vector3D v1, v2;

		kdtree->setNOfNeighbours(5);

		if (_use_mls)
			v2 = MLS.getPolyPosition();
		else
			v2 = MLS.getPlanePosition();

		kdtree->queryPosition(v2);
		
		for (int j=1; j<5; j++) {
			if (selector->isValid(kdtree->getNeighbourPositionIndex(j))) {
				radii += selector->getSurfel(kdtree->getNeighbourPositionIndex(j))->getRadius();
				v1 = kdtree->getNeighbourPosition(j);
				density += Vector3D::distance(v1,v2);
				count += 1.0f;
			} else
				while(0) {};
		}
		if (count == 0)
			continue;

		radii /= count;
		density /= count;
		
		if (_density_check) {
			// if too high density skip this surfel
			if (density < radii * 1.5f ) 
				continue;
		}

		if (_use_mls)
			_new_positions[index] = MLS.getPolyPosition();
		else
			_new_positions[index] = MLS.getPlanePosition();

		
		// add this surfel
		all_positions->push_back(_new_positions[index]);
		
		
		index_array[index] = index;
		_new_query_grid->insert(index_array+index, _new_positions[index]);
		

		// check orientation of normal
		Vector3D normal;
		Vector3D nearest_normal;
		Vector3D axis1, axis2;
		Vector3D result_pos, result_norm;
		float radius;
		QColor diffuse, specular;

		if (_use_mls)
			normal = MLS.getPolyNormal();
		else
			normal = MLS.getPlaneNormal();

		selector->getSurfelWorldNormal(kdtree->getNeighbourPositionIndex(0), nearest_normal);
	
		if (Vector3D::dotProduct(nearest_normal, normal) < 0)
			normal *= -1.0f;
		
		normal.getTwoOrthogonals(axis1, axis2);
		axis1.normalize();
		axis2.normalize();
		radius = radii;
		diffuse.setRgb(255,0,0);
		specular.setRgb(255,0,0);

		MtrMultVector3DHom (inv_trafo_matrix, _new_positions[index], &result_pos);
		MtrMultVector3D (inv_trafo_matrix, normal, &result_norm);

		surfel = Scene::getInstance()->getActiveObject()->getSurfelCollection()->addSurfel(
			result_pos,
			result_norm,
			axis1*radius,
			axis2*radius,
			radius,
			diffuse.rgb(),
			specular.rgb());

		
		sfls.push_back(surfel);
		poses.push_back(_new_positions[index]);
	}
	
	reColorSurfels(sfls,poses);
	
	selector->buildTree();
	
	_MCdelete((unsigned long)all_positions);
	delete all_positions;
	_MCdelete((unsigned long)index_array);
	delete[] index_array;

	QApplication::restoreOverrideCursor();

}




UndoAction* MLSScatterer::executeUndo() {
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	SurfelInterface *surfel;
	std::vector<SurfelInterface*> sfls;
	std::vector<Vector3D> poses;
	std::vector<Vector3D> *all_positions;
	MovingLeastSquares MLS(1e-3f, 30);
	unsigned int *index_array;
	unsigned i;
	unsigned index;
	float tmp_param;

	Scene *scene;
	Selector *selector;
	KdTree *kdtree;

	MyDataTypes::TransformationMatrix16f trafo_matrix, inv_trafo_matrix;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selector->buildTree();
	kdtree = selector->returnKdTree();

	if ( !scene->getActiveObject() || !kdtree) {
		QApplication::restoreOverrideCursor();
		return 0;
	}

	//get both transformationmatrix and its inverse
	scene->getActiveObject()->getTransformationMatrix(trafo_matrix);
	MtrInverse4x4f(trafo_matrix, inv_trafo_matrix);

	all_positions = selector->getAllPositionsWorldStdVec();
	index_array = new unsigned int[all_positions->size()];
	_MCnew((unsigned long)index_array, sizeof(index_array), __FILE__, __LINE__);
	
	kdtree->setNOfNeighbours(_neighbourhood_size);
	
	kdtree->queryPosition((*all_positions)[0]);
	tmp_param = kdtree->getSquaredDistance((unsigned int)(0.8f * (float)_neighbourhood_size));
	
	//tmp_param = sqrt(tmp_param);
	tmp_param = 10;
	
	// checking for error
	if (tmp_param == 0)
		while(0) {};

	MLS.setNOfNeighbours(_neighbourhood_size);
	MLS.setGaussParam(_param_mls);

	// set up query grid
	if (_new_query_grid) {
		_MCdelete((unsigned long)_new_query_grid);
		delete _new_query_grid;
	}

	_new_query_grid = new QueryGrid<uInt> (7, maxdiameter); // 7 used to be 5
	_MCnew( (unsigned long)_new_query_grid, sizeof(_new_query_grid), __FILE__, __LINE__);

	// sort points into pqueue
	doSorting();
	
	// first only plane projection
	for (i=0; i<_no_of_points; i++) {
		// get index with minimum length
		index = _distance_pqueue.getMinIndex();
		// remove index with minimum length
		_distance_pqueue.removeMin();
		
		// use this min for a MLS fit
		// rebuild mls after every 20 points
		if (i%20 == 0)
			MLS.setPositions(*all_positions);
		
		MLS.setSource(_new_positions[index]);

		if (_use_mls)
			MLS.doFullMLS();
		else
			MLS.doPlaneProjection();

		if (!MLS.isValid())
			continue;

		// check if we got a too high sampling density
		//  (not so sophisticated like in ~~> ParticleSimPlugin)
		float density = 0.f;
		float radii = 0.f;
		float count = 0.f;
		Vector3D v1, v2;

		kdtree->setNOfNeighbours(5);

		if (_use_mls)
			v2 = MLS.getPolyPosition();
		else
			v2 = MLS.getPlanePosition();

		kdtree->queryPosition(v2);
		
		for (int j=1; j<5; j++) {
			if (selector->isValid(kdtree->getNeighbourPositionIndex(j))) {
				radii += selector->getSurfel(kdtree->getNeighbourPositionIndex(j))->getRadius();
				v1 = kdtree->getNeighbourPosition(j);
				density += Vector3D::distance(v1,v2);
				count += 1.0f;
			} else
				while(0) {};
		}
		if (count == 0)
			continue;

		radii /= count;
		density /= count;
		
		if (_density_check) {
			// if too high density skip this surfel
			if (density < radii * 1.5f ) 
				continue;
		}
		
		if (_use_mls)
			_new_positions[index] = MLS.getPolyPosition();
		else
			_new_positions[index] = MLS.getPlanePosition();
		
		// add this surfel
		all_positions->push_back(_new_positions[index]);
		
		index_array[index] = index;
		_new_query_grid->insert(index_array+index, _new_positions[index]);
		

		// check orientation of normal
		Vector3D normal;
		Vector3D nearest_normal;
		Vector3D axis1, axis2;
		Vector3D result_pos, result_norm;
		float radius;
		QColor diffuse, specular;
		
		if (_use_mls)
			normal = MLS.getPolyNormal();
		else
			normal = MLS.getPlaneNormal();

		selector->getSurfelWorldNormal(kdtree->getNeighbourPositionIndex(0), nearest_normal);
	
		if (Vector3D::dotProduct(nearest_normal, normal) < 0)
			normal *= -1.0f;
		
		normal.getTwoOrthogonals(axis1, axis2);
		axis1.normalize();
		axis2.normalize();
		radius = radii;
		diffuse.setRgb(255,0,0);
		specular.setRgb(255,0,0);

		MtrMultVector3DHom (inv_trafo_matrix, _new_positions[index], &result_pos);
		MtrMultVector3D (inv_trafo_matrix, normal, &result_norm);
		surfel = Scene::getInstance()->getActiveObject()->getSurfelCollection()->addSurfel(
			result_pos,
			result_norm,
			axis1*radius,
			axis2*radius,
			radius,
			diffuse.rgb(),
			specular.rgb());

		
		sfls.push_back(surfel);
		poses.push_back(_new_positions[index]);
	}
	
	reColorSurfels(sfls,poses);
	
	selector->buildTree();

	// make undoaction

	InsertUndo *undo;
	std::map<unsigned int, Object*> tmp;
	tmp[sfls.size()] = Scene::getInstance()->getActiveObject();

	undo = new InsertUndo(sfls,tmp);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);

	_MCdelete((unsigned long)all_positions);
	delete all_positions;
	_MCdelete((unsigned long)index_array);
	delete[] index_array;

	QApplication::restoreOverrideCursor();

	return undo;

}


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
