// Title:   SurfelResizer.cpp
// Created: Thu Sep 25 15:14:59 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "SurfelResizer.h"

#include "../../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/Scene/src/Scene.h"

#include "Selector.h"

#include <qapplication.h>

#include "NewOverload.h"

SurfelResizer::SurfelResizer() : NormalAction() {
	_name = "resize surfels";
	_length = 512.;
	_retain_scale = false;
}

void SurfelResizer::analyse() {
	Scene *scene;
	Object *object;
	SurfelCollection *collection;
	SurfelInterface *surfel;
	Vector3D tmp;
	int n;
	float f;
	MyDataTypes::TransformationMatrix16f trafo_matrix;
	
	scene = Scene::getInstance();
	_bb_upper_corner = Vector3D(-1.e20f,-1.e20f,-1.e20f);
	_bb_lower_corner = Vector3D(1.e20f,1.e20f, 1.e20f);
	
	n = 0;

	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
		collection = object->getSurfelCollection();
		object->getTransformationMatrix(trafo_matrix);
	
		for (surfel = collection->getFirstSurfel(); surfel != 0; surfel = collection->getNextSurfel()) {
			// calculate bounding box			
			MtrMultVector3DHom(trafo_matrix,surfel->getPosition(),&tmp);
			if (tmp[0]<_bb_lower_corner[0])
				_bb_lower_corner[0] = tmp[0];
			if (tmp[1]<_bb_lower_corner[1])
				_bb_lower_corner[1] = tmp[1];
			if (tmp[2]<_bb_lower_corner[2])
				_bb_lower_corner[2] = tmp[2];
			if (tmp[0]>_bb_upper_corner[0])
				_bb_upper_corner[0] = tmp[0];
			if (tmp[1]>_bb_upper_corner[1])
				_bb_upper_corner[1] = tmp[1];
			if (tmp[2]>_bb_upper_corner[2])
				_bb_upper_corner[2] = tmp[2];
			// calculate center of gravity
			_center_of_gravity += tmp;
			n++;
		}

	}
	_center_of_gravity /= (float)n;
	
	// get the widest diameter
	f = 0;
	_diameter = 0;
	for (int i=0; i<3; i++) {
		f = _bb_upper_corner[i] - _bb_lower_corner[i];
		if (f > _diameter)
			_diameter = f;
	}
	
	_bb_center = (_bb_lower_corner + _bb_upper_corner) / 2.;
}

void SurfelResizer::resize() {
	Scene *scene;
	Object *object;
	SurfelCollection *collection;
	SurfelInterface *surfel;
	
	int propdesc;

	_scale = _length / _diameter;
	
	scene = Scene::getInstance();
	propdesc = SurfelInterface::getEditablePropertyDescriptor();

	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
		collection = object->getSurfelCollection();
		//object->setPosition(object->getPosition() * _scale);
		
		for (surfel = collection->getFirstSurfel(); surfel != 0; surfel = collection->getNextSurfel()) {
			
			scene->setSurfelPosition(surfel, surfel->getPosition() * _scale, false);
		
			if (propdesc & SurfelInterface::TANGENT_AXES)
				scene->setSurfelTangentAxes(surfel, surfel->getAxisOne() * _scale,surfel->getAxisTwo() * _scale, false);
			else
				scene->setSurfelRadius(surfel,surfel->getRadius() * _scale,false);
		}
	}


/*	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject())
		object->setPosition(object->getPosition() * _scale);
*/
	_bb_center *= _scale;
	_bb_lower_corner *= _scale;
	_bb_upper_corner *= _scale;
	_diameter *= _scale;
}
	

void SurfelResizer::translateAndScale() {
	Scene *scene;
	Object *object;
	
	scene = Scene::getInstance();
	// adjust the center
	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) 
		object->setPosition(object->getPosition() * _scale - _bb_center, false);
	
	
	if (_retain_scale)
		scene->scale(1./_scale,1./_scale,1./_scale);

/*
	MyDataTypes::cameraPosition cam;
	
	cam.scaleTranslationMatrix[0] = 1; cam.scaleTranslationMatrix[4] = 0; cam.scaleTranslationMatrix[ 8] = 0; cam.scaleTranslationMatrix[12] = 0;
	cam.scaleTranslationMatrix[1] = 0; cam.scaleTranslationMatrix[5] = 1; cam.scaleTranslationMatrix[ 9] = 0; cam.scaleTranslationMatrix[13] = 0;
	cam.scaleTranslationMatrix[2] = 0; cam.scaleTranslationMatrix[6] = 0; cam.scaleTranslationMatrix[10] = 1; cam.scaleTranslationMatrix[14] = 10;
	cam.scaleTranslationMatrix[3] = 0; cam.scaleTranslationMatrix[7] = 0; cam.scaleTranslationMatrix[11] = 0; cam.scaleTranslationMatrix[15] = 1;

	cam.rotationMatrix[0] = 1; cam.rotationMatrix[4] = 0; cam.rotationMatrix[ 8] = 0; cam.rotationMatrix[12] = 0;
	cam.rotationMatrix[1] = 0; cam.rotationMatrix[5] = 1; cam.rotationMatrix[ 9] = 0; cam.rotationMatrix[13] = 0;
	cam.rotationMatrix[2] = 0; cam.rotationMatrix[6] = 0; cam.rotationMatrix[10] = 1; cam.rotationMatrix[14] = 0;
	cam.rotationMatrix[3] = 0; cam.rotationMatrix[7] = 0; cam.rotationMatrix[11] = 0; cam.rotationMatrix[15] = 1;

	scene->setCameraPosition(cam);
*/	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

}

void SurfelResizer::execute() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	analyse();
	resize();
	translateAndScale();
	Selector::getInstance()->buildTree();

	QApplication::restoreOverrideCursor();
}

UndoAction* SurfelResizer::executeUndo() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	analyse();
	resize();
	translateAndScale();
	Selector::getInstance()->buildTree();

	QApplication::restoreOverrideCursor();

	UndoAction *undo = new SurfelResizerUndo(_scale,_bb_center);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);

	return undo;
}

/**
 * undo
 */

SurfelResizerUndo::SurfelResizerUndo(float scale, const Vector3D& translation) {
	_scale = scale;
	_translation = translation;
}

UndoAction* SurfelResizerUndo::undo() {
	Scene *scene;
	Object *object;
	SurfelCollection *collection;
	SurfelInterface *surfel;
	int propdesc;

	scene = Scene::getInstance();
	propdesc = SurfelInterface::getEditablePropertyDescriptor();

	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) {
		collection = object->getSurfelCollection();
			
		for (surfel = collection->getFirstSurfel(); surfel != 0; surfel = collection->getNextSurfel()) {
			scene->setSurfelPosition(surfel, surfel->getPosition() / _scale, false);
			if (propdesc & SurfelInterface::TANGENT_AXES)
				scene->setSurfelTangentAxes(surfel, surfel->getAxisOne() / _scale,surfel->getAxisTwo() * _scale, false);
			else
				scene->setSurfelRadius(surfel,surfel->getRadius() / _scale,false);
		}
		
	}
	
	// adjust the center
	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) 
		object->setPosition(object->getPosition() / _scale + _translation, false);
	
	scene->scale(_scale,_scale,_scale);
	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	Selector::getInstance()->buildTree();

	UndoAction *undo = new SurfelResizerUndo(1.f / _scale, _translation * -1.f);
	_MCnew( (unsigned long) undo, sizeof(undo), __FILE__, __LINE__);

	return undo;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
